/*
 
  copyright (c) 2004, Michal Salaban <emes@pld-linux.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License, version 2 as
  published by the Free Software Foundation (see file COPYING for details).

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
*/

/* $Id: parseopt.c,v 1.3 2004/01/29 22:49:20 emes Exp $ */

#include "parseopt.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

/*
 * helper functions
 */

void * _parseopt_alloc (size_t size)
{
    void    *res = (void *) malloc (size);
    if (res != NULL) return res;

    fprintf (stderr, "malloc() failed, exiting...");
    exit (-1);
}

void * parseopt_error (char *error)
{
    fprintf (stderr, "Invalid option: ");
    fprintf (stderr, error);
    fprintf (stderr, "\n");
}

char * valptr (char *pat, char *arg)
{
    /* compare unless characters don't match or we hit string's end */
    while ( (*pat == *arg) &&
            (*pat != '\0') &&
            (*arg != '\0') ) {
        pat++;
        arg++;
    }

    /* this is an option string */
    if ((*pat == '\0') && (*arg == '=')) return arg + 1;
    else return NULL;
}

int * parse_bool (char *str)
{
    char    *on[] = {
                "on",
                "yes",
                "1",
                "enabled",
                NULL
            },
            *off[] = {
                "off",
                "no",
                "0",
                "disabled",
                NULL
            };
    int     i;
    int     *res = (int *) _parseopt_alloc (sizeof (int));

    i = 0;
    while (on[i] != NULL)
        if (strcasecmp (on[i], str) == 0) {
            *res = 1;
            return res;
        } else i++;

    i = 0;
    while (off[i] != NULL)
        if (strcasecmp (off[i], str) == 0) {
            *res = 0;
            return res;
        } else i++;
    
    free (res);
    return NULL;
}

double * parse_double (char *str)
{
    double   *res = (double *) _parseopt_alloc (sizeof (double));
    char    *endptr;

    *res = strtod (str, &endptr);
    if ((*endptr == '\0') && (! errno)) return res;

    free (res);
    return NULL;

}

long int * parse_int (char *str)
{
    long int    *res = (long int *) _parseopt_alloc (sizeof (int));
    char        *endptr;

    *res = strtol (str, &endptr, 10);
    if ((*endptr == '\0') && (! errno)) return res;

    free (res);
    return NULL;
}

long int * parse_size (char *str)
{
    long int    *res = (long int *) _parseopt_alloc (sizeof (int));
    long int    mult;
    char        *endptr;

    *res = strtol (str, &endptr, 10);
    if (errno) return NULL;

    /* size in bytes */
    if (*endptr == '\0') mult = 1;
    /* kilobytes */
    else if (*endptr == 'k') mult = 1024;
    /* megabytes and so on... */
    else if (*endptr == 'M') mult = 1024 * 1024;
    else if (*endptr == 'G') mult = 1024 * 1024 * 1024;
    /*
     * following ones cause overflow
    else if (*endptr == 'T') mult = 1024 * 1024 * 1024 * 1024;
    else if (*endptr == 'P') mult = 1024 * 1024 * 1024 * 1024 * 1024;
     */
    else return NULL;

    *res *= mult;
    return res;
}

void * parseval (int otype, char *val)
{
    switch (otype) {
        case O_BOOL:
                return (void *) parse_bool (val);
        case O_INT:
                return (void *) parse_int (val);
        case O_DOUBLE:
                return (void *) parse_double (val);
        case O_SIZE:
                return (void *) parse_size (val);
        case O_STRING:
                return (void *) strcpy (
                                    (char *) _parseopt_alloc (strlen (val)),
                                    val
                                  );
    }
    return NULL;
}

void ** get_val_ptr (
        optval      (*vals)[],
        int         lastopt,
        int         otype )
{
    int     i = 0;

    while ((*vals)[i].selector != lastopt)
        if ((*vals)[i].selector == otype) return &((*vals)[i].val);
        else i++;
}


/* get_opt ()
 * 
 * Get option from arg string, using options representation table opts
 * and no-option constant lastopt
 *
 * Return value:
 *  optval consisting of option selector and pointer to option value,
 *  NULL pointer value indicates parse error
 *
 * Warning:
 *  - lastopt must be present in optrepr table and should be the last entry,
 */
optval get_opt (
        optrepr     (*opts)[],
        int         lastopt,
        char        *arg )
{
    int         i = 0;
    char        *val;
    optval      res;
    
    while ((*opts)[i].selector != lastopt) {
        val = valptr ((*opts)[i].repr, arg);
        /* we got it? */
        if (val != NULL) break;
        else i++;
    }

    res.selector = (*opts)[i].selector;
    if (val != NULL) res.val = parseval ((*opts)[i].otype, val);
    else res.val = NULL;

    return res;
}

/* put_opt ()
 *
 * Fills appropriate field in vals table using option selector in osel,
 * option value in oval and no-option selector in lastopt.
 *
 * Return value:
 *  none
 *
 * Warning:
 *  - lastopt must be present in vals table and should be the last entry,
 */
void put_opt (
        optval      (*vals)[],
        int         lastopt,
        int         osel,
        void        *oval )
{
    void    **where = get_val_ptr (vals, lastopt, osel);
   
    *where = oval;
}


void * get_val (
        optval      (*vals)[],
        int         lastopt,
        int         osel )
{
    return *(get_val_ptr (vals, lastopt, osel));
}    


int fetch_opt_array (
        optval      (*vals)[],
        optrepr     (*reprs)[],
        int         lastopt,
        int         argc,
        char        *(*argv)[] )
{
    int     i;
    optval  v;

    for (i = 0; i < argc; i++) {
        v = get_opt (reprs, lastopt, (*argv)[i]);
        if (v.val == NULL) {
            parseopt_error ((*argv)[i]);
            return 0;
        }
        if (v.selector != lastopt)
            put_opt (vals, lastopt, v.selector, v.val);
    }
    
    return 1;
}

/* $Id: parseopt.c,v 1.3 2004/01/29 22:49:20 emes Exp $ */
